@@ -1,13 +1,11 @@
-from .datasets import (
-    Dataset,
-    EmbeddingColumnNames,
-    ExampleDatasets,
-    Schema,
-    load_example,
-)
-from .session import Session, active_session, close_app, launch_app
+from .datasets.dataset import Dataset
+from .datasets.fixtures import ExampleDatasets, load_example
+from .datasets.schema import EmbeddingColumnNames, RetrievalEmbeddingColumnNames, Schema
+from .session.session import Session, active_session, close_app, launch_app
+from .trace.fixtures import load_example_traces
+from .trace.trace_dataset import TraceDataset
 
-__version__ = "0.0.17"
+__version__ = "1.2.0"
 
 # module level doc-string
 __doc__ = """
@@ -28,6 +26,7 @@ Here are just a few of the things that phoenix does well:
 __all__ = [
     "Dataset",
     "EmbeddingColumnNames",
+    "RetrievalEmbeddingColumnNames",
     "Schema",
     "load_example",
     "ExampleDatasets",
@@ -35,4 +34,6 @@ __all__ = [
     "close_app",
     "launch_app",
     "Session",
+    "load_example_traces",
+    "TraceDataset",
 ]
