@@ -1,12 +1,14 @@
 import json
 import math
 import re
+import threading
 from abc import ABC, abstractmethod
 from collections import defaultdict
+from contextlib import contextmanager
 from dataclasses import dataclass, field, fields, replace
 from datetime import datetime, timedelta, timezone
-from enum import IntEnum, auto, unique
-from functools import cached_property, lru_cache
+from enum import Enum, IntEnum, auto, unique
+from functools import cached_property
 from itertools import chain, groupby, repeat, starmap
 from random import random
 from typing import (
@@ -14,6 +16,8 @@ from typing import (
     BinaryIO,
     Callable,
     Dict,
+    Generic,
+    Hashable,
     Iterable,
     Iterator,
     List,
@@ -33,6 +37,7 @@ from uuid import uuid4
 from weakref import ProxyType, proxy
 
 import numpy as np
+import numpy.typing as npt
 import pandas as pd
 from pandas.core.dtypes.common import (
     is_bool_dtype,
@@ -43,6 +48,9 @@ from pandas.core.dtypes.common import (
 from typing_extensions import TypeAlias, TypeGuard
 from wrapt import ObjectProxy
 
+from phoenix.config import GENERATED_DATASET_NAME_PREFIX
+from phoenix.datetime_utils import floor_to_minute
+
 
 class DimensionRole(IntEnum):
     ...
@@ -164,7 +172,22 @@ class Embedding(CompositeDimensionSpec):
                 yield value
 
 
-class DatasetRole(IntEnum):
+@dataclass(frozen=True)
+class RetrievalEmbedding(Embedding):
+    context_retrieval_ids: Optional[str] = None
+    context_retrieval_scores: Optional[str] = None
+
+    def __iter__(self) -> Iterator[str]:
+        yield from super().__iter__()
+        for value in (
+            self.context_retrieval_ids,
+            self.context_retrieval_scores,
+        ):
+            if isinstance(value, str) and value:
+                yield value
+
+
+class DatasetRole(Enum):
     """A dataframe's role in a Model: primary or reference (as
     baseline for drift).
     """
@@ -177,10 +200,86 @@ PRIMARY = DatasetRole.PRIMARY
 REFERENCE = DatasetRole.REFERENCE
 
 
-@lru_cache(maxsize=len(DatasetRole))
-def _series_nan(length: int) -> "pd.Series[float]":
-    """Useful as a substitute for a non-existent column."""
-    return pd.Series(np.full(length, float("nan")))
+@dataclass(frozen=True, repr=False, eq=False)
+class _ConstantValueSeriesFactory:
+    """The intent is to share memory for readonly situations where a constant
+    value, e.g. NaN, is expected from a non-existent dataframe column by the
+    downstream calculations. To avoid duplicating memory when we require such
+    columns multiple times or need to subset them frequently, we first check for
+    a previously allocated array that's longer, and subset from it, thereby
+    sharing the memory already allocated (since the value is constant and does
+    not change). If the cached array is not long enough, we then allocate and
+    cache a new array, replacing the old one. This cache is intended to be
+    attached to a Model instance, and will be garbage collected alongside it,
+    while the `lru_cache` decorator's cache is scoped at the module level and
+    can potentially leak memory.
+    """
+
+    value: Any = field(default=np.nan)
+    _dtype: Any = field(init=False, default=None)
+    _cached_array: npt.NDArray[Any] = field(
+        init=False,
+        default_factory=lambda: np.empty(0),
+    )
+    """If a longer Series is requested, the cached array is expanded;
+    otherwise, a subset can be returned, assuming it won't be altered by the
+    caller.
+    """
+    _lock: threading.Lock = field(
+        init=False,
+        default_factory=threading.Lock,
+    )
+    """A lock is applied at the class instance level for thread safety, with
+    minimal overhead expected, unless too many callers simultaneously rely on
+    the same instance.
+    """
+
+    def __post_init__(self) -> None:
+        if isinstance(self.value, float) and math.isnan(self.value):
+            object.__setattr__(self, "_dtype", np.float32)
+
+    def __call__(self, length: int) -> "pd.Series[Any]":
+        with self._lock:
+            if length > len(self._cached_array):
+                object.__setattr__(
+                    self,
+                    "_cached_array",
+                    np.full(length, self.value, dtype=self._dtype),
+                )
+            return pd.Series(self._cached_array[:length])
+
+
+_Key = TypeVar("_Key", bound=Hashable)
+_Value = TypeVar("_Value")
+
+
+@dataclass(frozen=True, repr=False, eq=False)
+class _Cache(Generic[_Key, _Value]):
+    """A thread-safe type-safe generic cache backed by a dictionary.
+
+    Example
+    -------
+    >>> c = _Cache[str, int]()
+    >>> with c() as cache:
+    >>>     cache["1"] = 2
+    >>> with c() as cache:
+    >>>     print(cache["1"])
+    2
+    """
+
+    _cache: Dict[_Key, _Value] = field(
+        init=False,
+        default_factory=dict,
+    )
+    _lock: threading.Lock = field(
+        init=False,
+        default_factory=threading.Lock,
+    )
+
+    @contextmanager
+    def __call__(self) -> Iterator[Dict[_Key, _Value]]:
+        with self._lock:
+            yield self._cache
 
 
 DataFrameOrSeries: TypeAlias = Union[pd.DataFrame, "pd.Series[Any]"]
@@ -192,8 +291,8 @@ are extracted (instead of getting a column back)."""
 @dataclass(frozen=True)
 class Column:
     """Extracts a value (i.e. scalar) from pd.Series or a series (i.e. a
-    column) from pd.DataFrame. If not found, returns NaN or a series of NaNs,
-    respectively.
+    column) from pd.DataFrame. If not found, return the default value
+    (e.g. NaN) or a series of the default value (on each row), respectively.
     """
 
     name: str = ""
@@ -202,6 +301,9 @@ class Column:
     can remain relatively declarative and compact, i.e. by not having to keep
     writing `if ... is not None:` everywhere. Dummy columns always return NaNs,
     because they have random column names not found in any dataframe."""
+    _default: _ConstantValueSeriesFactory = field(
+        default_factory=_ConstantValueSeriesFactory,
+    )
 
     def __post_init__(self) -> None:
         if not self.name:
@@ -218,25 +320,29 @@ class Column:
 
     def __call__(self, data: DataFrameOrSeries) -> Any:
         """Extracts a value from series, or a series from a dataframe. If
-        not found, return NaN or a series of NaNs, respectively.
+        not found, return the default value (e.g. NaN) or a series of the
+        default value (on each row), respectively.
         """
         if isinstance(data, pd.DataFrame):
             try:
                 return data.loc[:, self.name]
             except KeyError:
-                # It's important to glue the index to the NaN series, so it
-                # would look like the series came from the dataframe.
-                return _series_nan(len(data)).set_axis(data.index)
+                # It's important to glue the index to the default series,
+                # so it would look like the series came from the dataframe.
+                return self._default(len(data)).set_axis(data.index)
         if isinstance(data, pd.Series):
             try:
                 return data.at[self.name]
             except KeyError:
-                return float("nan")
-        raise ValueError("invalid data: %s" % repr(data))
+                return self._default.value
+        raise ValueError(f"invalid data: {repr(data)}")
 
     def __iter__(self) -> Iterator[str]:
         """This is to partake in the iteration of column names by a
-        larger data structure of which this object is a member.
+        larger data structure of which this object is a member. Dummy
+        columns need not be yielded because they represent columns
+        that don't (physically) exist (i.e. they are just fillers for
+        the Model to make the Schema whole).
         """
         if not self.is_dummy:
             yield self.name
@@ -305,16 +411,16 @@ class ScalarDimension(Dimension):
     @property
     def min_max(self) -> Tuple[Any, Any]:
         if self._model is None:
-            return float("nan"), float("nan")
+            return np.nan, np.nan
         model = cast(Model, self._model)
         return model.dimension_min_max_from_all_df(self.name)
 
     @property
-    def categories(self) -> Tuple[Any, ...]:
+    def categories(self) -> Tuple[str, ...]:
         if self._model is None or self.data_type is CONTINUOUS:
             return ()
         model = cast(Model, self._model)
-        return tuple(model.dimension_categories_from_all_df(self.name).dropna())
+        return model.dimension_categories_from_all_datasets(self.name)
 
 
 @dataclass(frozen=True)
@@ -327,12 +433,24 @@ class EmbeddingDimension(Dimension):
         super().__post_init__()
         if not self.display_name:
             object.__setattr__(self, "display_name", self.name)
+        # optimization: share the default series factory
+        object.__setattr__(
+            self,
+            "link_to_data",
+            replace(self.link_to_data, _default=self._default),
+        )
+        # optimization: share the default series factory
+        object.__setattr__(
+            self,
+            "raw_data",
+            replace(self.raw_data, _default=self._default),
+        )
 
     @classmethod
-    def from_(cls, emb: Embedding, **kwargs: Any) -> "EmbeddingDimension":
-        """Use `from_` instead of `__init__` because the latter is needed by
-        replace() and we don't want to clobber the generated version.
-        """
+    def from_embedding(cls, emb: Embedding, **kwargs: Any) -> "EmbeddingDimension":
+        # Use `from_embedding` instead of `__init__` because the latter is
+        # needed by replace() and we don't want to clobber the generated
+        # version.
         return cls(
             _coerce_str(emb.vector),
             link_to_data=Column(_coerce_str(emb.link_to_data)),
@@ -352,6 +470,39 @@ class EmbeddingDimension(Dimension):
             yield from self.link_to_data
 
 
+@dataclass(frozen=True)
+class RetrievalEmbeddingDimension(EmbeddingDimension):
+    context_retrieval_ids: Column = field(default_factory=Column)
+    context_retrieval_scores: Column = field(default_factory=Column)
+
+    @classmethod
+    def from_prompt_embedding(
+        cls, emb: RetrievalEmbedding, **kwargs: Any
+    ) -> "RetrievalEmbeddingDimension":
+        # Use `from_embedding` instead of `__init__` because the latter is
+        # needed by replace() and we don't want to clobber the generated
+        # version.
+        return cls(
+            _coerce_str(emb.vector),
+            link_to_data=Column(_coerce_str(emb.link_to_data)),
+            raw_data=Column(_coerce_str(emb.raw_data)),
+            display_name=_coerce_str(emb.display_name),
+            context_retrieval_ids=Column(_coerce_str(emb.context_retrieval_ids)),
+            context_retrieval_scores=Column(_coerce_str(emb.context_retrieval_scores)),
+            **kwargs,
+        )
+
+    def __iter__(self) -> Iterator[str]:
+        """This is to partake in the iteration of column names by a
+        larger data structure of which this object is a member.
+        """
+        yield from super().__iter__()
+        if not self.context_retrieval_ids.is_dummy:
+            yield from self.context_retrieval_ids
+        if not self.context_retrieval_scores.is_dummy:
+            yield from self.context_retrieval_scores
+
+
 Name: TypeAlias = str
 ColumnKey: TypeAlias = Union[Name, Column, SingularDimensionalRole]
 MultiDimensionKey: TypeAlias = Union[MultiDimensionalRole, Sequence[DimensionRole]]
@@ -460,7 +611,7 @@ class Event(ModelData):
         return self._self_id
 
     def null_value(self) -> float:
-        return float("nan")
+        return np.nan
 
     @overload
     def __getitem__(self, key: ColumnKey) -> Any:
@@ -494,10 +645,8 @@ class Events(ModelData):
     @cached_property
     def time_range(self) -> TimeRange:
         if self._self_model is None or self.empty:
-            # NOTE: as of Python 3.8.16, pandas 1.5.3:
-            # >>> isinstance(pd.NaT, datetime.datetime)
-            # True
-            return TimeRange(pd.NaT, pd.NaT)  # type: ignore
+            now = datetime.now(timezone.utc)
+            return TimeRange(now, now)
         model = cast(Model, self._self_model)
         min_max = _agg_min_max(model[TIMESTAMP](self))
         start_time = cast(datetime, min_max.min())
@@ -506,8 +655,8 @@ class Events(ModelData):
         # open and one minute is the smallest interval allowed.
         stop_time = end_time + timedelta(minutes=1)
         # Round down to the nearest minute.
-        start = _floor_to_minute(start_time)
-        stop = _floor_to_minute(stop_time)
+        start = floor_to_minute(start_time)
+        stop = floor_to_minute(stop_time)
         return TimeRange(start, stop)
 
     def __iter__(self) -> Iterator[Event]:
@@ -553,10 +702,31 @@ class Dataset(Events):
     def name(self) -> str:
         return self._self_name
 
+    @property
+    def display_name(self) -> str:
+        """
+        Return a human-readable name for this dataset. If the user doesn't
+        provide a name, the name is generated but this name is not human
+        friendly. Falls back to the role of the dataset if no name is provided.
+        """
+        ds_name = self._self_name
+        if ds_name.startswith(GENERATED_DATASET_NAME_PREFIX):
+            # The generated names are UUIDs so use the role as the name
+            return "primary" if self.role is DatasetRole.PRIMARY else "reference"
+        return ds_name
+
     @property
     def role(self) -> DatasetRole:
         return self._self_role
 
+    @property
+    def is_empty(self) -> bool:
+        return len(self) == 0
+
+    @cached_property
+    def primary_key(self) -> pd.Index:
+        return pd.Index(self[PREDICTION_ID])
+
     @overload
     def __getitem__(self, key: ColumnKey) -> "pd.Series[Any]":
         ...
@@ -591,6 +761,10 @@ class Model:
     _dimensions: Dict[Name, Dimension]
     _dim_names_by_role: Dict[DimensionRole, List[Name]]
     _original_columns_by_role: Dict[DatasetRole, pd.Index]
+    _default_timestamps_factory: _ConstantValueSeriesFactory
+    _nan_series_factory: _ConstantValueSeriesFactory
+    _dimension_categories_from_all_datasets: _Cache[Name, Tuple[str, ...]]
+    _dimension_min_max_from_all_datasets: _Cache[Name, Tuple[float, float]]
 
     def __init__(
         self,
@@ -603,6 +777,18 @@ class Model:
         # TODO: Consider moving validations here.
         df_already_validated: bool = False,
     ):
+        # memoization
+        object.__setattr__(
+            self,
+            "_dimension_categories_from_all_datasets",
+            _Cache[Name, "pd.Series[Any]"](),
+        )
+        object.__setattr__(
+            self,
+            "_dimension_min_max_from_all_datasets",
+            _Cache[Name, Tuple[float, float]](),
+        )
+
         df_names, dfs = cast(
             Tuple[Iterable[Name], Iterable[pd.DataFrame]],
             zip(*_coerce_tuple(dataframes)),
@@ -627,6 +813,17 @@ class Model:
             {role: dataset.columns for role, dataset in self._datasets.items()},
         )
 
+        object.__setattr__(
+            self,
+            "_nan_series_factory",
+            _ConstantValueSeriesFactory(np.nan),
+        )
+        object.__setattr__(
+            self,
+            "_default_timestamps_factory",
+            _ConstantValueSeriesFactory(datetime.now(timezone.utc)),
+        )
+
         # Store dimensions by name. In general, a dimension's name is that of
         # its primary column. But dimensions with special roles, e.g.
         # `PREDICTION_ID` have names that match their roles, e.g.
@@ -656,11 +853,18 @@ class Model:
         # Guess the data type, i.e. continuous or discrete, for scalar
         # features and tags.
         for dim in self[ScalarDimension]:
-            if dim.data_type is UNKNOWN:
-                data_type = _guess_data_type(map(dim, self._datasets.values()))
-                self._dimensions[dim.name] = replace(dim, data_type=data_type)
-
-        default_timestamp = datetime.now(timezone.utc)
+            if dim.data_type is not UNKNOWN:
+                continue
+            self._dimensions[dim.name] = replace(
+                dim,
+                data_type=(
+                    _guess_data_type(
+                        dataset.loc[:, dim.name]
+                        for dataset in self._datasets.values()
+                        if dim.name in dataset.columns
+                    )
+                ),
+            )
 
         # Add PREDICTION_ID if missing.
         # Add TIMESTAMP if missing.
@@ -686,7 +890,7 @@ class Model:
                 self._new_dimension(TIMESTAMP),
             )
             if dim_time.name not in df_original_columns:
-                df[dim_time.name] = _series_constant(len(df), default_timestamp)
+                df[dim_time.name] = self._default_timestamps_factory(len(df))
             else:
                 if not timestamps_already_normalized:
                     # Don't clobber the original (or any other column).
@@ -708,28 +912,52 @@ class Model:
                 df, name=dataset.name, role=dataset_role
             )
 
+    @cached_property
+    def is_empty(self) -> bool:
+        """Returns True if the model has no data."""
+        return not any(map(len, self._datasets.values()))
+
     def export_rows_as_parquet_file(
         self,
-        rows: Mapping[DatasetRole, Iterable[int]],
+        row_numbers: Mapping[DatasetRole, Iterable[int]],
         parquet_file: BinaryIO,
+        cluster_ids: Optional[Mapping[DatasetRole, Mapping[int, str]]] = None,
     ) -> None:
         """
         Given row numbers, exports dataframe subset into parquet file.
-        Duplicate rows are removed.
+        Duplicate rows are removed. If the model hase more than one dataset, a
+        new column is added to the dataframe containing the dataset name of
+        each row in the exported data. The name of the added column will be
+        `__phoenix_dataset_name__`.
 
         Parameters
         ----------
-        rows: Mapping[DatasetRole, Iterable[int]]
-            mapping of dataset type to list of row numbers
+        row_numbers: Mapping[DatasetRole, Iterable[int]]
+            mapping of dataset role to list of row numbers
         parquet_file: file handle
             output parquet file handle
+        cluster_ids: Optional[Mapping[DatasetRole, Mapping[int, str]]]
+            mapping of dataset role to mapping of row number to cluster id.
+            If cluster_ids is non-empty, a new column is inserted to the
+            dataframe containing the cluster IDs of each row in the exported
+            data. The name of the added column name is `__phoenix_cluster_id__`.
         """
-        pd.concat(
-            self[dataset_role][sorted(row_numbers)].loc[
-                :, self._original_columns_by_role[dataset_role]
+        export_dataframes = [pd.DataFrame()]
+        model_has_multiple_datasets = sum(not df.empty for df in self._datasets.values()) > 1
+        for dataset_role, numbers in row_numbers.items():
+            df = self._datasets[dataset_role]
+            columns = [
+                df.columns.get_loc(column_name)
+                for column_name in self._original_columns_by_role[dataset_role]
             ]
-            for dataset_role, row_numbers in rows.items()
-        ).to_parquet(
+            rows = pd.Series(sorted(set(numbers)))
+            filtered_df = df.iloc[rows, columns].reset_index(drop=True)
+            if model_has_multiple_datasets:
+                filtered_df["__phoenix_dataset_name__"] = df.display_name
+            if cluster_ids and (ids := cluster_ids.get(dataset_role)):
+                filtered_df["__phoenix_cluster_id__"] = rows.apply(ids.get)
+            export_dataframes.append(filtered_df)
+        pd.concat(export_dataframes).to_parquet(
             parquet_file,
             index=False,
             allow_truncated_timestamps=True,
@@ -745,7 +973,7 @@ class Model:
             dim
             for dim in self._dimensions.values()
             if not dim.is_dummy
-            and dim.role not in (PREDICTION_ID, TIMESTAMP)
+            and dim.role not in (PREDICTION_ID, TIMESTAMP, RESPONSE)
             and isinstance(dim, ScalarDimension)
         )
 
@@ -760,18 +988,46 @@ class Model:
             if not dim.is_dummy and isinstance(dim, EmbeddingDimension)
         )
 
-    def dimension_categories_from_all_df(self, dimension_name: Name) -> "pd.Series[Any]":
+    def dimension_categories_from_all_datasets(
+        self,
+        dimension_name: Name,
+    ) -> Tuple[str, ...]:
         dim = self[dimension_name]
-        categories_by_df = (dim[role].unique() for role in DatasetRole)
-        all_values_combined = chain.from_iterable(categories_by_df)
-        return pd.Series(all_values_combined).sort_values().drop_duplicates()
+        if dim.data_type is CONTINUOUS:
+            return cast(Tuple[str, ...], ())
+        with self._dimension_categories_from_all_datasets() as cache:
+            try:
+                return cache[dimension_name]
+            except KeyError:
+                pass
+        categories_by_dataset = (
+            pd.Series(dim[role].unique()).dropna().astype(str) for role in DatasetRole
+        )
+        all_values_combined = chain.from_iterable(categories_by_dataset)
+        ans = tuple(np.sort(pd.Series(all_values_combined).unique()))
+        with self._dimension_categories_from_all_datasets() as cache:
+            cache[dimension_name] = ans
+        return ans
 
-    def dimension_min_max_from_all_df(self, dimension_name: Name) -> Tuple[Any, Any]:
+    def dimension_min_max_from_all_df(
+        self,
+        dimension_name: Name,
+    ) -> Tuple[float, float]:
         dim = self[dimension_name]
+        if dim.data_type is not CONTINUOUS:
+            return (np.nan, np.nan)
+        with self._dimension_min_max_from_all_datasets() as cache:
+            try:
+                return cache[dimension_name]
+            except KeyError:
+                pass
         min_max_by_df = (_agg_min_max(dim[df_role]) for df_role in DatasetRole)
         all_values_combined = chain.from_iterable(min_max_by_df)
         min_max = _agg_min_max(pd.Series(all_values_combined))
-        return min_max.min(), min_max.max()
+        ans = (min_max.min(), min_max.max())
+        with self._dimension_min_max_from_all_datasets() as cache:
+            cache[dimension_name] = ans
+        return ans
 
     @overload
     def __getitem__(self, key: Type[Dataset]) -> Iterator[Dataset]:
@@ -823,9 +1079,15 @@ class Model:
         if _is_dimension_type_filter(key):
             return self._get_multi_dims_by_type(key)
         if key is ScalarDimension:
-            return filter(lambda dim: type(dim) is ScalarDimension, self._dimensions.values())
+            return filter(
+                lambda dim: type(dim) is ScalarDimension,
+                self._dimensions.values(),
+            )
         if key is EmbeddingDimension:
-            return filter(lambda dim: type(dim) is EmbeddingDimension, self._dimensions.values())
+            return filter(
+                lambda dim: type(dim) is EmbeddingDimension,
+                self._dimensions.values(),
+            )
         if key is Dimension:
             return self._dimensions.values()
         raise KeyError(f"invalid key: {repr(key)}")
@@ -886,14 +1148,30 @@ class Model:
         self, obj: Any, cls: Type[Dimension] = ScalarDimension, **kwargs: Any
     ) -> Dimension:
         """Creates a new Dimension or copies an existing one, setting the
-        model weak reference to the `self` Model instance.
+        model weak reference to the `self` Model instance, and sharing the
+        NaN series factory as an optimization.
         """
         if isinstance(obj, Name):
-            return cls(obj, **kwargs, _model=proxy(self))
+            return cls(
+                obj,
+                **kwargs,
+                _model=proxy(self),
+                _default=self._nan_series_factory,
+            )
         if isinstance(obj, DimensionRole):
-            return cls(role=obj, **kwargs, _model=proxy(self))
+            return cls(
+                role=obj,
+                **kwargs,
+                _model=proxy(self),
+                _default=self._nan_series_factory,
+            )
         if isinstance(obj, Dimension):
-            return replace(obj, **kwargs, _model=proxy(self))
+            return replace(
+                obj,
+                **kwargs,
+                _model=proxy(self),
+                _default=self._nan_series_factory,
+            )
         raise ValueError(f"invalid argument: {repr(obj)}")
 
     def _new_dataset(
@@ -917,8 +1195,8 @@ class Schema(SchemaSpec):
     prediction_score: Optional[str] = None
     actual_label: Optional[str] = None
     actual_score: Optional[str] = None
-    prompt: Optional[Embedding] = None
-    response: Optional[Embedding] = None
+    prompt: Optional[RetrievalEmbedding] = None
+    response: Optional[Union[str, Embedding]] = None
     features: Iterable[Union[str, CompositeDimensionSpec]] = field(default_factory=list)
     tags: Iterable[Union[str, CompositeDimensionSpec]] = field(default_factory=list)
 
@@ -943,7 +1221,7 @@ class Schema(SchemaSpec):
         with dummy dimensions for ones omitted by user. The dummy dimensions
         have randomly generated names that can change for each iteration, but
         currently there's no need to iterate more than once."""
-        for spec, role, data_type in chain(
+        for spec, dimension_role, data_type in chain(
             (
                 (self.prediction_id, PREDICTION_ID, DISCRETE),
                 (self.timestamp, TIMESTAMP, CONTINUOUS),
@@ -959,16 +1237,34 @@ class Schema(SchemaSpec):
         ):
             if not isinstance(spec, CompositeDimensionSpec):
                 spec = _coerce_str(spec)
-            assert isinstance(role, DimensionRole)  # for mypy
+            assert isinstance(dimension_role, DimensionRole)  # for mypy
             if isinstance(spec, str):
-                if role in (PROMPT, RESPONSE):
-                    yield EmbeddingDimension(spec, role=role, data_type=data_type)
+                if dimension_role is PROMPT:
+                    yield RetrievalEmbeddingDimension(
+                        spec,
+                        role=dimension_role,
+                        data_type=data_type,
+                    )
                 else:
-                    yield ScalarDimension(spec, role=role, data_type=data_type)
+                    yield ScalarDimension(
+                        spec,
+                        role=dimension_role,
+                        data_type=data_type,
+                    )
+            elif isinstance(spec, RetrievalEmbedding):
+                yield RetrievalEmbeddingDimension.from_prompt_embedding(
+                    spec,
+                    role=dimension_role,
+                    data_type=data_type,
+                )
             elif isinstance(spec, Embedding):
-                yield EmbeddingDimension.from_(spec, role=role, data_type=data_type)
+                yield EmbeddingDimension.from_embedding(
+                    spec,
+                    role=dimension_role,
+                    data_type=data_type,
+                )
             else:
-                raise TypeError(f"{role} has unrecognized type: {type(spec)}")
+                raise TypeError(f"{dimension_role} has unrecognized type: {type(spec)}")
 
     def __call__(
         self,
@@ -1065,12 +1361,6 @@ def _iterate_except_str(obj: Any) -> Iterator[Any]:
         yield obj
 
 
-@lru_cache(maxsize=len(DatasetRole))
-# Note that NaN can't be cached, because NaN != NaN.
-def _series_constant(length: int, constant: Any) -> "pd.Series[Any]":
-    return pd.Series(np.full(length, constant))
-
-
 def _series_uuid(length: int) -> "pd.Series[str]":
     return pd.Series(map(lambda _: uuid4(), range(length)))
 
@@ -1100,7 +1390,13 @@ def _coerce_tuple(
 def _coerce_str_column_names(
     dataframes: Iterable[pd.DataFrame],
 ) -> Iterator[pd.DataFrame]:
-    return (df.set_axis(df.columns.astype(str), axis=1) for df in dataframes)
+    return (
+        df.set_axis(
+            df.columns.astype(str),
+            axis=1,
+        )
+        for df in dataframes
+    )
 
 
 _T = TypeVar("_T")
@@ -1139,32 +1435,6 @@ def _title_case_no_underscore(name: str) -> str:
     return _id_pat.sub("ID", name.replace("_", " ").title())
 
 
-MINUTE_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:00%z"
-
-
-def _floor_to_minute(dt: datetime) -> datetime:
-    """Floor datetime to the minute by taking a round-trip through string
-    format because there isn't always an available function to strip the
-    nanoseconds if present."""
-    try:
-        dt_as_string = dt.astimezone(
-            timezone.utc,
-        ).strftime(
-            MINUTE_DATETIME_FORMAT,
-        )
-    except ValueError:
-        # NOTE: as of Python 3.8.16, pandas 1.5.3:
-        # >>> isinstance(pd.NaT, datetime.datetime)
-        # True
-        return cast(datetime, pd.NaT)
-    return datetime.strptime(
-        dt_as_string,
-        MINUTE_DATETIME_FORMAT,
-    ).astimezone(
-        timezone.utc,
-    )
-
-
 def _jsonify(obj: Any) -> Any:
     if getattr(obj, "__dataclass_fields__", None):
         return {
@@ -1197,3 +1467,4 @@ def _objectify(json_data: Any) -> Any:
         except TypeError:
             pass
     raise ValueError(f"invalid json data: {repr(json_data)}")
+    raise ValueError(f"invalid json data: {repr(json_data)}")
