@@ -1,6 +1,6 @@
 import logging
 from pathlib import Path
-from typing import Optional, Union
+from typing import Any, NamedTuple, Optional, Union
 
 from starlette.applications import Starlette
 from starlette.datastructures import QueryParams
@@ -9,26 +9,60 @@ from starlette.exceptions import HTTPException
 from starlette.middleware import Middleware
 from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
 from starlette.requests import Request
-from starlette.responses import FileResponse, Response
+from starlette.responses import FileResponse, PlainTextResponse, Response
 from starlette.routing import Mount, Route, WebSocketRoute
 from starlette.staticfiles import StaticFiles
+from starlette.templating import Jinja2Templates
 from starlette.types import Scope
 from starlette.websockets import WebSocket
 from strawberry.asgi import GraphQL
 from strawberry.schema import BaseSchema
 
+import phoenix
 from phoenix.config import SERVER_DIR
 from phoenix.core.model_schema import Model
-
-from .api.context import Context
-from .api.schema import schema
+from phoenix.core.traces import Traces
+from phoenix.pointcloud.umap_parameters import UMAPParameters
+from phoenix.server.api.context import Context
+from phoenix.server.api.schema import schema
+from phoenix.server.span_handler import SpanHandler
 
 logger = logging.getLogger(__name__)
 
+templates = Jinja2Templates(directory=SERVER_DIR / "templates")
+
+
+class AppConfig(NamedTuple):
+    has_corpus: bool
+    min_dist: float
+    n_neighbors: int
+    n_samples: int
+
+
+def _get_basename(request: Request) -> str:
+    """
+    Determine the basename for the API and static content.
+    If the server is running in a hosted notebook, the base URL
+    must be configured to point to the hosted notebook's proxy.
+    """
+    is_sagemaker = "sagemaker" in str(request.url)
+    if is_sagemaker:
+        # Sagemaker sets up a proxy at /proxy/6006
+        # NB: this is the only port that is open
+        return "/proxy/6006"
+    # URL is relative to / for colab and local
+    return ""
+
 
 class Static(StaticFiles):
     "Static file serving with a fallback to index.html"
 
+    _app_config: AppConfig
+
+    def __init__(self, *, app_config: AppConfig, **kwargs: Any):
+        self._app_config = app_config
+        super().__init__(**kwargs)
+
     async def get_response(self, path: str, scope: Scope) -> Response:
         response = None
         try:
@@ -37,10 +71,19 @@ class Static(StaticFiles):
             if e.status_code != 404:
                 raise e
             # Fallback to to the index.html
-            full_path, stat_result = self.lookup_path("index.html")
-            if stat_result is None:
-                raise RuntimeError("Failed to find index.html")
-            response = self.file_response(full_path, stat_result, scope)
+            request = Request(scope)
+
+            response = templates.TemplateResponse(
+                "index.html",
+                context={
+                    "has_corpus": self._app_config.has_corpus,
+                    "min_dist": self._app_config.min_dist,
+                    "n_neighbors": self._app_config.n_neighbors,
+                    "n_samples": self._app_config.n_samples,
+                    "basename": _get_basename(request),
+                    "request": request,
+                },
+            )
         except Exception as e:
             raise e
         return response
@@ -57,15 +100,19 @@ class HeadersMiddleware(BaseHTTPMiddleware):
         return response
 
 
-class GraphQLWithContext(GraphQL):
+class GraphQLWithContext(GraphQL):  # type: ignore
     def __init__(
         self,
         schema: BaseSchema,
         model: Model,
         export_path: Path,
         graphiql: bool = False,
+        corpus: Optional[Model] = None,
+        traces: Optional[Traces] = None,
     ) -> None:
         self.model = model
+        self.corpus = corpus
+        self.traces = traces
         self.export_path = export_path
         super().__init__(schema, graphiql=graphiql)
 
@@ -78,6 +125,8 @@ class GraphQLWithContext(GraphQL):
             request=request,
             response=response,
             model=self.model,
+            corpus=self.corpus,
+            traces=self.traces,
             export_path=self.export_path,
         )
 
@@ -97,14 +146,23 @@ class Download(HTTPEndpoint):
         )
 
 
+async def version(_: Request) -> PlainTextResponse:
+    return PlainTextResponse(f"{phoenix.__version__}")
+
+
 def create_app(
     export_path: Path,
     model: Model,
+    umap_params: UMAPParameters,
+    corpus: Optional[Model] = None,
+    traces: Optional[Traces] = None,
     debug: bool = False,
 ) -> Starlette:
     graphql = GraphQLWithContext(
         schema=schema,
         model=model,
+        corpus=corpus,
+        traces=traces,
         export_path=export_path,
         graphiql=True,
     )
@@ -113,7 +171,22 @@ def create_app(
             Middleware(HeadersMiddleware),
         ],
         debug=debug,
-        routes=[
+        routes=(
+            []
+            if traces is None
+            else [
+                Route(
+                    "/v1/spans",
+                    type(
+                        "SpanEndpoint",
+                        (SpanHandler,),
+                        {"queue": traces},
+                    ),
+                ),
+            ]
+        )
+        + [
+            Route("/arize_phoenix_version", version),
             Route(
                 "/exports",
                 type(
@@ -131,7 +204,12 @@ def create_app(
                 "/",
                 app=Static(
                     directory=SERVER_DIR / "static",
-                    html=True,
+                    app_config=AppConfig(
+                        has_corpus=corpus is not None,
+                        min_dist=umap_params.min_dist,
+                        n_neighbors=umap_params.n_neighbors,
+                        n_samples=umap_params.n_samples,
+                    ),
                 ),
                 name="static",
             ),
